"""Models
------

Consumers for the dasf-broker-django app.
"""

# Disclaimer
# ----------
#
# Copyright (C) 2022 Helmholtz-Zentrum Hereon
#
# This file is part of dasf-broker-django and is released under the
# EUPL-1.2 license.
# See LICENSE in the root of the repository for full licensing details.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the EUROPEAN UNION PUBLIC LICENCE v. 1.2 or later
# as published by the European Commission.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# EUPL-1.2 license for more details.
#
# You should have received a copy of the EUPL-1.2 license along with this
# program. If not, see https://www.eupl.eu/.

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Optional
from uuid import uuid4

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer
from django.utils.functional import cached_property
from guardian.shortcuts import (
    assign_perm,
    get_anonymous_user,
    get_users_with_perms,
)

from dasf_broker import app_settings
from dasf_broker.constants import PropertyKeys

if TYPE_CHECKING:
    from dasf_broker.models import DASFTopic


class TopicProducer(JsonWebsocketConsumer):
    """A producer of messages for a certain topic."""

    @cached_property
    def dasf_topic(self) -> Optional[DASFTopic]:
        from dasf_broker.models import DASFTopic

        return DASFTopic.objects.filter(slug=self.dasf_topic_slug).first()

    def connect(self):
        user = self.scope["user"]
        self.user = get_anonymous_user() if user.is_anonymous else user
        self.dasf_topic_slug = self.scope["url_route"]["kwargs"]["slug"]

        self.accept()

    def receive_json(self, content: Dict):
        """Distibute the message to the consumers."""
        from dasf_broker.models import DASFTopic

        # create a topic for the response
        props = content.get("properties", {})

        # dummy messageId (for now)
        message_id = content.setdefault("messageId", str(uuid4()))
        message_id = str(uuid4())

        ack = {
            "messageId": message_id,
        }

        if "context" in content:
            ack["context"] = content["context"]

        topic = self.dasf_topic

        if topic is None:
            if (
                app_settings.DASF_CREATE_TOPIC_ON_MESSAGE
                and self.user.has_perm("dasf_broker.add_dasftopic")
            ):
                topic = DASFTopic.objects.create(
                    slug=self.dasf_topic_slug, consumer=self.user
                )
                assign_perm("can_produce", self.user, topic)
            elif not app_settings.DASF_CREATE_TOPIC_ON_MESSAGE:
                ack["result"] = f"Topic {self.dasf_topic_slug} does not exist."
            else:
                ack["result"] = "User is not allowed to create topics."

        if (
            topic is not None
            and topic.is_public
            or self.user.has_perm("dasf_broker.can_produce", topic)
        ):
            # create a topic for the response

            if PropertyKeys.RESPONSE_TOPIC in props:
                response_topic, created = DASFTopic.objects.get_or_create(
                    slug=props[PropertyKeys.RESPONSE_TOPIC],
                    consumer=self.user,
                    is_response_topic=True,
                )
                if created:
                    for user in get_users_with_perms(
                        topic, "dasf_broker.can_consume"
                    ):
                        assign_perm("can_produce", user, response_topic)

            async_to_sync(self.channel_layer.group_send)(
                f"dasf_topic_{self.dasf_topic_slug}",
                {"type": "dasf.message", "content": content},
            )
            # acknowledge the message
            ack["result"] = "ok"
            self.send_json(ack)
        elif topic is not None:

            ack["result"] = "User is not allowed to publish to this topic"
        self.send_json(ack)


class TopicConsumer(JsonWebsocketConsumer):
    """A consumer of messages."""

    @property
    def dasf_topic(self) -> Optional[DASFTopic]:
        from dasf_broker.models import DASFTopic

        return DASFTopic.objects.filter(slug=self.dasf_topic_slug).first()

    def connect(self):
        user = self.scope["user"]
        self.user = get_anonymous_user() if user.is_anonymous else user
        self.dasf_topic_slug = slug = self.scope["url_route"]["kwargs"]["slug"]
        self.room_group_name = f"dasf_topic_{slug}"

        # Join room group
        async_to_sync(self.channel_layer.group_add)(
            self.room_group_name, self.channel_name
        )

        self.accept()

    def receive_json(self, content, **kwargs):
        # delete response topic from database
        topic = self.dasf_topic
        if (
            topic is not None
            and topic.is_public
            or self.user.has_perm("dasf_broker.can_produce", topic)
        ):
            if topic.is_response_topic:
                topic.delete()

    def disconnect(self, close_code):
        # Leave room group
        async_to_sync(self.channel_layer.group_discard)(
            self.room_group_name, self.channel_name
        )

    def dasf_message(self, event):
        topic = self.dasf_topic
        if topic is not None and self.user.has_perm("can_consume", topic):
            self.send_json(event["content"])
